
#include "cerberus.h"
#include "entities.h"


/*!
 * 
 */
bool cEntityBase::createEntity( eEntity entityType )
{
  // entity must be an element or boulder
  if( (getType() != eElement) && (getType() != eBoulder) )
  	return false;

	// is this entity below another one? 
  if( getNextEntity() )
    return false;

	// if it's an element, it must be flat    
  if( (getType() == eElement) && !(((cElement*)this)->isFlat()) )
  		return false;
  
  cEntity* entity;
  switch( entityType ) {
  case eTree:
    {
      entity = new cEntity( eTree, mWorld, mDevice );
      scene::IAnimatedMeshSceneNode* Node=mDevice->getSceneManager()->addAnimatedMeshSceneNode( mWorld->getTreeMesh(), mWorld->getNode() );
      if( Node ) {
        Node->setMaterialFlag( video::EMF_LIGHTING, false );
        Node->setPosition( core::vector3df(mIndexI+0.5, mGroundLevel+mHeight, mIndexJ+0.5) );
        Node->setRotation( core::vector3df(0.0, mWorld->getRandomNumber()*360.0, 0.0) );
        entity->setNode( Node );
        entity->setIndex( mIndexI, mIndexJ );
        entity->setHeight( 1.767767 );
        entity->setGroundLevel( mGroundLevel+mHeight );
        setNextEntity( entity );
        entity->setPreviousEntity( this );
      } else {
        fprintf( stdout, "couldn't allocate tree" );
      }
    }
    break;
  case eBoulder:
    {
      entity = new cEntity( eBoulder, mWorld, mDevice );
      scene::IAnimatedMeshSceneNode* Node=mDevice->getSceneManager()->addAnimatedMeshSceneNode( mWorld->getBoulderMesh(), mWorld->getNode() );
      if( Node ) {
        Node->setMaterialFlag( video::EMF_LIGHTING, false );
        Node->setPosition( core::vector3df(mIndexI+0.5, mGroundLevel+mHeight, mIndexJ+0.5) );
        Node->setRotation( core::vector3df(0.0, mWorld->getRandomNumber()*360.0, 0.0) );
        entity->setNode( Node );
        entity->setIndex( mIndexI, mIndexJ );
        entity->setHeight( 1.0/sqrt(2.0) );
        entity->setGroundLevel( mGroundLevel+mHeight );
        setNextEntity( entity );
        entity->setPreviousEntity( this );
				scene::ITriangleSelector* entitySelector =
									mWorld->getSceneManager()->createTriangleSelector( mWorld->getBoulderMesh(), Node );
				Node->setTriangleSelector( entitySelector );
				entitySelector->drop();
      } else {
        fprintf( stdout, "couldn't allocate tree" );
      }
    }
    break;
  case eRobot:
    {
      entity = new cEntity( eRobot, mWorld, mDevice );
      scene::IAnimatedMeshSceneNode* Node=mDevice->getSceneManager()->addAnimatedMeshSceneNode( mWorld->getRobotMesh(), mWorld->getNode() );
      if( Node ) {
				Node->setMaterialFlag( video::EMF_LIGHTING, false );
				Node->setPosition( core::vector3df(mIndexI+0.5, mGroundLevel+mHeight, mIndexJ+0.5) );
				core::vector3df diffVector=mWorld->getSoul()->getPosition()-Node->getPosition();
				Node->setRotation( core::vector3df(0.0, 180.0-atan2( diffVector.Z, diffVector.X )*180.0/core::PI, 0.0) );
        entity->setNode( Node );
        entity->setIndex( mIndexI, mIndexJ );
        entity->setHeight( 1.072501 );
        entity->setGroundLevel( mGroundLevel+mHeight );
        setNextEntity( entity );
        entity->setPreviousEntity( this );
      } else {
        fprintf( stdout, "couldn't allocate robot" );
      }
    }
    break;
  default:
    return NULL;
  }
	
	return true;
}


int cEntityBase::getEnergy()
{
	int energy;
	
	switch( mType )
	{
	case eTree: energy = eTreeNrg; break;
	case eBoulder: energy = eBoulderNrg; break;
	case eRobot: energy = eRobotNrg; break;
	default: energy = 0; break;
	}
	
	return energy;
}

/**
 * Setting variables.
 */
cEntity::cEntity( eEntity type, cWorld* world, IrrlichtDevice* device ) :
	cEntityBase( type, world, device ), mNode( NULL )
{
}


/**
 * Remove node from scene manager
 */
cEntity::~cEntity()
{
	if( mNode )
		mNode->remove();
}


/**
 * Setting variables and initialie object.
 */
cElement::cElement( cWorld* world, IrrlichtDevice* device ) : cEntityBase( eElement, world, device )
{
	setIndex( 0, 0 );
  mLevel=0;
  mHeight = 0;
  mGroundLevel = 0.0;

 	mNode=NULL;
  mFlat=false;
  
  mLogger = mDevice->getLogger();
}


/**
 * Free entities on this element and remove node.
 */
cElement::~cElement()
{
	// remove stack of entities on this element
	cEntityBase* entity = getNextEntity();
	while( entity ) {
		cEntityBase* temp = entity->getNextEntity();
		delete entity;
		entity = temp;
	}
	setNextEntity( NULL );
	
	if( mNode )
		mNode->remove();
}


/*!
 * 
 */
//void cElement::rotateRobot()
//{
  //if( mRobot ) {
  //  scene::IAnimatedMeshSceneNode* Node=mRobot->getNode();
  //  core::vector3df diffVector=mWorld->getSoul()->getPosition()-Node->getPosition();
  //  Node->setRotation( core::vector3df(0.0, 180-atan2( diffVector.Z, diffVector.X )*180.0/core::PI, 0.0) );
  //}
//}


/*!
 * Initialize soul and camera.
 *
 * @param camera : pointer to the camera node, of which the soul takes care of
 */
cSoul::cSoul( scene::ICameraSceneNode* camera ) : mCamera( camera ), mTheta( core::PI/2.0 ), 
        mThetaMin( 2.0f/180.0f*core::PI ), mThetaMax( 178.0f/180.0f*core::PI )
{
	// camera settings
  mRotateSpeed=1.0f;
  mCamera->setNearValue( 0.1f );
  mCamera->setFarValue( 12000.0f );
  
  // internal settings
  mEnergy = 7;
  
  // soul always looks to the middle of the landscape
  mPhi=atan2( (double)(16-mIndexJ), (double)(16-mIndexI) );
  
	// initialize View
  updateView();
}


/*!
 * Update view (direction where we looking at).
 */
void cSoul::updateView()
{
	mCamera->setPosition( core::vector3df( mIndexI+0.5, mGroundLevel+1.0, mIndexJ+0.5 ) );
	mCamera->setTarget( mCamera->getPosition() +
		core::vector3df(cos(mPhi)*sin(mTheta), cos(mTheta), sin(mPhi)*sin(mTheta)) );            
	mCamera->setUpVector( core::vector3df(-cos(mPhi)*cos(mTheta), sin(mTheta), -sin(mPhi)*cos(mTheta)) );            
}


/*!
 * 
 */
void cSoul::setTarget( core::vector3df target )
{
	core::vector3df diffVector=target-getPosition();

	mPhi=atan2( diffVector.Z, diffVector.X );
	mTheta=atan2( sqrt(diffVector.X*diffVector.X+diffVector.Z*diffVector.Z), diffVector.Y );
}


/*!
 * Change view (direction where we looking at).
 */
void cSoul::changeView( f32 px, f32 py )
{
	// determine new rotation angles
	mPhi -= px*mRotateSpeed;
	mTheta += py*mRotateSpeed;
	if( mTheta<mThetaMin ) mTheta=mThetaMin;
	if( mTheta>mThetaMax ) mTheta=mThetaMax;
	if( mPhi<0.0 ) 				mPhi+=2.0f*core::PI;
	if( mPhi>2*core::PI )  mPhi-=2.0f*core::PI;
	
	updateView();
}


/*!
 * Quick turn around.
 */
void cSoul::turnAround()
{
	// turn around 180 degrees
	mPhi += core::PI;
	if( mPhi>2*core::PI )  mPhi-=2.0f*core::PI;
	mTheta = core::PI/2;
	
	updateView();
}


/**
 * Substract an amount of energy from the souls energy.
 * If the energy of the soul becomes 0 or lower return
 * false - the soul dies of exhaustion.
 *
 * @param sub : amount of energy to substract
 * @return false if energy exhausted, true otherwise
 */
bool cSoul::substractFromEnergy( int sub )
{ 
  mEnergy -= sub;

  if( mEnergy<=0 )
    return false;
  
  return true;
}
